# Copyright (c) 2009-2012 testtools developers. See LICENSE for details.

"""Test result objects."""

__all__ = [
    "CopyStreamResult",
    "DetailsDict",
    "ExtendedToOriginalDecorator",
    "ExtendedToStreamDecorator",
    "MultiTestResult",
    "ResourcedToStreamDecorator",
    "StreamFailFast",
    "StreamResult",
    "StreamResultRouter",
    "StreamSummary",
    "StreamTagger",
    "StreamToDict",
    "StreamToExtendedDecorator",
    "StreamToQueue",
    "Tagger",
    "TestByTestResult",
    "TestControl",
    "TestResult",
    "TestResultDecorator",
    "TextTestResult",
    "ThreadsafeForwardingResult",
    "TimestampingStreamResult",
]

from testtools.testresult.real import (
    CopyStreamResult,
    DetailsDict,
    ExtendedToOriginalDecorator,
    ExtendedToStreamDecorator,
    MultiTestResult,
    ResourcedToStreamDecorator,
    StreamFailFast,
    StreamResult,
    StreamResultRouter,
    StreamSummary,
    StreamTagger,
    StreamToDict,
    StreamToExtendedDecorator,
    StreamToQueue,
    Tagger,
    TestByTestResult,
    TestControl,
    TestResult,
    TestResultDecorator,
    TextTestResult,
    ThreadsafeForwardingResult,
    TimestampingStreamResult,
)
